import flask

from ...contracts_generated import AbstraLibApiEditorEnvVarsModel
from ...repositories import env_vars_repository
from ...usage import usage


def get_editor_bp():
    bp = flask.Blueprint("editor_env_vars", __name__)

    # 1s pooling in this route
    @bp.get("/")
    def _get_env_vars():
        return [env_var.to_dict() for env_var in env_vars_repository.list()]

    @bp.post("/")
    @usage
    def _set_env_var():
        if flask.request.json is None:
            flask.abort(400)
        env_var = env_vars_repository.set(
            AbstraLibApiEditorEnvVarsModel.from_dict(flask.request.json)
        )
        return env_var.to_dict()

    @bp.delete("/<key>")
    @usage
    def _unset_env_var(key: str):
        env_var = env_vars_repository.unset(
            AbstraLibApiEditorEnvVarsModel(key=key, value="")
        )
        return env_var.to_dict()

    return bp



from ..widget_base import Input
from typing import Any


class CustomInput(Input):
    type = 'custom-input'

    def __init__(self, key: str, html_body: str, **kwargs):
        super().__init__(key)
        self.set_props(dict(html_body=html_body, **kwargs))

    def set_props(self, props):
        self.html_body = props.get('html_body', '')
        self.label = props.get('label', '')
        self.html_head = props.get('html_head', '')
        self.height = props.get('height', 200)
        self.css = props.get('css', '')
        self.js = props.get('js', '')
        self.value = props.get('initial_value', None)
        self.full_width = props.get('full_width', False)

    def serialize_value(self) ->Any:
        return self.value

    def render(self, ctx: dict):
        return {'type': self.type, 'key': self.key, 'label': self.label,
            'value': self.serialize_value(), 'htmlBody': self.html_body,
            'height': self.height, 'htmlHead': self.html_head, 'css': self.
            css, 'js': self.js, 'fullWidth': self.full_width}

    def parse_value(self, value) ->Any:
        return value



from unittest import TestCase

from tests.fixtures import clear_dir, init_dir

from ...server.controller.main import MainController
from .missing_packages_in_requirements import MissingPackagesInRequirements


class MissingPackagesInRequirementsTest(TestCase):
    def setUp(self) -> None:
        self.root = init_dir()

    def tearDown(self) -> None:
        clear_dir(self.root)

    def test_missing_packages_in_requirements_valid_default(self):
        requirements_file = self.root / "requirements.txt"
        self.assertFalse(requirements_file.exists())
        rule = MissingPackagesInRequirements()
        self.assertEqual(len(rule.find_issues()), 0)

    def test_missing_packages_in_requirements_valid_with_packages(self):
        requirements_file = self.root / "requirements.txt"
        requirements_file.touch()
        requirements_file.write_text("pandas==1.0.0")
        constroller = MainController()
        script = constroller.create_script("New script", "script.py")
        code = "import pandas"
        script.file_path.write_text(code)
        rule = MissingPackagesInRequirements()
        self.assertEqual(len(rule.find_issues()), 0)

    def test_missing_packages_in_requirements_valid_with_packages_using_from_syntax(
        self,
    ):
        requirements_file = self.root / "requirements.txt"
        requirements_file.touch()
        requirements_file.write_text("pandas==1.0.0")
        constroller = MainController()
        script = constroller.create_script("New script", "script.py")
        code = "from pandas import foo"
        script.file_path.write_text(code)
        rule = MissingPackagesInRequirements()
        self.assertEqual(len(rule.find_issues()), 0)

    def test_missing_packages_in_requirements_invalid_requirements_file(self):
        constroller = MainController()
        script = constroller.create_script("New script", "script.py")
        code = "import pandas"
        script.file_path.write_text(code)
        rule = MissingPackagesInRequirements()
        self.assertEqual(len(rule.find_issues()), 1)

    def test_missing_packages_in_requirements_invalid_requirements_file_using_from_syntax(
        self,
    ):
        constroller = MainController()
        script = constroller.create_script("New script", "script.py")
        code = "from pandas import foo"
        script.file_path.write_text(code)
        rule = MissingPackagesInRequirements()
        self.assertEqual(len(rule.find_issues()), 1)

    def test_missing_packages_in_requirements_invalid_missing_package(self):
        requirements_file = self.root / "requirements.txt"
        requirements_file.touch()
        constroller = MainController()
        script = constroller.create_script("New script", "script.py")
        code = "import pandas"
        script.file_path.write_text(code)
        rule = MissingPackagesInRequirements()
        self.assertEqual(len(rule.find_issues()), 1)

    def test_missing_packages_with_submodule_import(self):
        requirements_file = self.root / "requirements.txt"
        requirements_file.touch()
        constroller = MainController()
        script = constroller.create_script("New script", "script.py")
        code = "import pandas.plotting"
        script.file_path.write_text(code)
        rule = MissingPackagesInRequirements()
        self.assertEqual(len(rule.find_issues()), 1)

    def test_missing_package_with_submodule_using_from_syntax(self):
        requirements_file = self.root / "requirements.txt"
        requirements_file.touch()
        constroller = MainController()
        script = constroller.create_script("New script", "script.py")
        code = "from pandas.plotting import foo"
        script.file_path.write_text(code)
        rule = MissingPackagesInRequirements()
        self.assertEqual(len(rule.find_issues()), 1)



import unittest

from tests.fixtures import clear_dir, get_cloud_client, init_dir


class TestPlayerApi(unittest.TestCase):
    def setUp(self) -> None:
        self.root = init_dir()
        self.client = get_cloud_client()

    def tearDown(self) -> None:
        clear_dir(self.root)

    def test_version(self):
        res = self.client.get("/_version")
        self.assertEqual(res.status_code, 200)
        self.assertEqual(res.text, "dev")



import flask

from ...usage import usage
from ...utils import is_it_true
from .main import MainController


def get_editor_bp(controller: MainController):
    bp = flask.Blueprint("editor_forms", __name__)

    @bp.get("/")
    @usage
    def _get_forms():
        return [f.editor_dto for f in controller.get_forms()]

    @bp.get("/<path:id>")
    @usage
    def _get_form(id: str):
        form = controller.get_form(id)
        if not form:
            flask.abort(404)
        return form.editor_dto

    @bp.delete("/<path:id>")
    @usage
    def _delete_form(id: str):
        remove_file = flask.request.args.get(
            "remove_file", default=False, type=is_it_true
        )
        controller.delete_form(id, remove_file)
        return {"success": True}

    @bp.post("/")
    @usage
    def _create_form():
        data = flask.request.json
        if not data:
            flask.abort(400)
        title = data.get("title")
        file = data.get("file")
        if not title or not file:
            flask.abort(400)
        workflow_position = data.get("position", (0, 0))
        form = controller.create_form(title, file, workflow_position)
        return form.editor_dto

    @bp.put("/<path:id>")
    @usage
    def _update_form(id: str):
        data = flask.request.json
        if not data:
            flask.abort(400)

        form = controller.update_stage(id, data)
        return form.editor_dto if form else None

    return bp



from unittest import TestCase

from tests.fixtures import clear_dir, init_dir

from .constants import get_persistent_dir
from .utils.dot_abstra import PERSISTENT_DIR


class GetPersistentDirTest(TestCase):
    def setUp(self) -> None:
        self.root = init_dir()

    def tearDown(self) -> None:
        clear_dir(self.root)

    def test_value(self):
        self.assertFalse((self.root / PERSISTENT_DIR).exists())
        self.assertEqual(self.root / PERSISTENT_DIR, get_persistent_dir())
        self.assertTrue((self.root / PERSISTENT_DIR).exists())



from typing import Optional

from .execution import ABSTRA_RUN_KEY, Execution, ResponseData


class HookExecution(Execution):
    @staticmethod
    def get_current_hook_execution() -> "HookExecution":
        execution = Execution.get_current_execution()
        if isinstance(execution, HookExecution):
            return execution

        raise Exception("No hook execution found")

    def handle_start(self) -> None:
        pass

    def handle_success(self) -> None:
        self.context.response = self.context.response or ("", 200, {})

    def handle_failure(self, e: Exception) -> None:
        self.context.response = self.context.response or ("", 500, {})

    def handle_lock_failed(self) -> None:
        self.context.response = self.context.response or ("", 409, {})

    def get_response(self) -> ResponseData:
        if self.context.response is None:
            raise Exception("No response set")

        return self.context.response

    def received_stage_run_id(self) -> Optional[str]:
        return self.context.request.query_params.get(ABSTRA_RUN_KEY)



from typing import Optional

from ..repositories.execution import ExecutionRepository
from ..repositories.execution_logs import ExecutionLogsRepository
from ..repositories.project.project import ScriptStage
from .execution import Execution, RequestData
from .stage_run_manager import StageRunManager


class ScriptExecution(Execution):
    stage_run_id: str

    def __init__(
        self,
        stage: ScriptStage,
        stage_run_id: str,
        stage_run_manager: StageRunManager,
        execution_repository: ExecutionRepository,
        execution_logs_repository: ExecutionLogsRepository,
    ):
        request = RequestData(
            headers={},
            body="",
            method="GET",
            query_params={},
        )

        self.stage_run_id = stage_run_id

        super().__init__(
            stage=stage,
            is_initial=True,
            request=request,
            execution_repository=execution_repository,
            execution_logs_repository=execution_logs_repository,
            stage_run_manager=stage_run_manager,
        )

    def handle_start(self) -> None:
        pass

    def handle_success(self) -> None:
        pass

    def handle_failure(self, e: Exception) -> None:
        pass

    def handle_lock_failed(self) -> None:
        pass

    def received_stage_run_id(self) -> Optional[str]:
        return self.stage_run_id



from .NumberInput import NumberInput
from .TagInput import TagInput
from .DropdownInput import DropdownInput
from .RatingInput import RatingInput
from .CurrencyInput import CurrencyInput
from .CardsInput import CardsInput
from .CnpjInput import CnpjInput
from .ListInput import ListInput
from .PlotlyOutput import PlotlyOutput
from .MultipleChoiceInput import MultipleChoiceInput
from .TextareaInput import TextareaInput
from .CheckboxInput import CheckboxInput
from .TextOutput import TextOutput
from .NumberSliderInput import NumberSliderInput
from .LinkOutput import LinkOutput
from .IframeOutput import IframeOutput
from .CodeInput import CodeInput
from .FileOutput import FileOutput
from .ImageOutput import ImageOutput
from .PhoneInput import PhoneInput
from .FileInput import FileInput
from .DateInput import DateInput
from .CpfInput import CpfInput
from .NpsInput import NpsInput
from .TimeInput import TimeInput
from .PandasOutput import PandasOutput
from .VideoInput import VideoInput
from .TextInput import TextInput
from .ChecklistInput import ChecklistInput
from .PandasRowSelectionInput import PandasRowSelectionInput
from .CustomInput import CustomInput
from .LatexOutput import LatexOutput
from .CameraInput import CameraInput
from .PasswordInput import PasswordInput
from .HtmlOutput import HtmlOutput
from .EmailInput import EmailInput
from .RichTextInput import RichTextInput
from .ToggleInput import ToggleInput
from .MarkdownOutput import MarkdownOutput
from .ImageInput import ImageInput
from .ProgressOutput import ProgressOutput




# flake8: noqa
from typing import List, Union, Dict, Any, TypedDict
import io

from ..reactive import Reactive
from abstra_internals.widgets.library import *
from abstra_internals.widgets import Input, Output
from abstra_internals.widgets.types import PlotlyFigure, PandasDataFrame, AbstraOption


class WidgetSchema:

  def __init__(self):
    self.widgets: List = []

  def __get_next_result_key(self):
    result_widgets = len(
        [w for w in self.widgets if getattr(w, "key", None) == "result"]
    )
    return "result" if result_widgets == 0 else f"result_{result_widgets}"

  def reactive(self, callback):
    self.widgets.append(Reactive(callback))
    return self

  def has_errors(self):
    return any([widget.has_errors() for widget in self.widgets])

  def set_values(self, values: Dict):
    for widget in self.widgets:
        if isinstance(widget, Output):
            continue
        elif isinstance(widget, Reactive):
            widget.set_value(values)
        elif widget.key in values:
            widget.set_value(values[widget.key], set_errors=True)


  def set_errors(self):
    for widget in self.widgets:
        if isinstance(widget, Output):
            continue
        widget.set_errors()

  def parse_value(self, form_answers: Dict) -> Dict:
    """Convert the answer from the form to the expected format
    Args:
        answer: The answer from the form
    Returns:
        The converted answer
    """
    answer: Dict = {}
    inputs = self.widgets
    for input in inputs:
        if isinstance(input, Reactive):
            answer = {**answer, **input.parse_value(form_answers)}
        elif isinstance(input, Input):
            answer[input.key] = input.parse_value(form_answers.get(input.key))
    return answer

  def get_input_widgets(self):
    return list(filter(lambda widget: isinstance(widget, Input), self.widgets))

  def render(self, context=None):
    output = []
    for widget in self.widgets:
        rendered_widget = widget.render(context)
        if isinstance(rendered_widget, list):
            output.extend(rendered_widget)
        else:
            output.append(rendered_widget)
    return output

  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  
  def read_camera(self, label: str, **kwargs):
    
    """Captures and uploads a photo using the user's camera.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (str): The initial value to display to the user. Defaults to None.
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          A dict containing the picture taken by the user FileResponse(file: TemporaryFile, url: str)
    """

    
  
    
    key = kwargs.pop("key", label)
  
    self.widgets.append(CameraInput(key, label, **kwargs))
    return self

    
  def read_cards(self, label: str, options: list, **kwargs):
    
    """Presents options as cards for the user. Cards may be selectable.

        Position Args: 
		label (str): The text related to this input
		options (list): List of dicts representing the cards, each dict can have the following keys: title, subtitle, image, description, topLeftExtra, topRightExtra
		
        Keyword Args: 
		multiple (bool): Whether the user can select multiple options. Defaults to False.
		initial_value (list): The initial value to display to the user. Defaults to None.
		searchable (bool): Whether to show a search bar. Defaults to False.
		layout (str): Whether the cards layout should be 'list' or 'grid'. Defaults to 'list'. The 'grid' only applies to desktop resolutions.
		columns (int): When layout is 'grid', how many columns to display
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The options/option selected by the user
    """

    
  
    
    key = kwargs.pop("key", label)
  
    self.widgets.append(CardsInput(key, label, options, **kwargs))
    return self

    
  def read_checkbox(self, label: str, **kwargs):
    
    """Enables a selection through a single checkbox interaction.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (str): The initial value to display to the user.
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The value entered by the user
    """

    
  
    
    key = kwargs.pop("key", label)
  
    self.widgets.append(CheckboxInput(key, label, **kwargs))
    return self

    
  def read_checklist(self, label: str, options: List[AbstraOption], **kwargs):
    
    """Displays a checklist allowing multiple item selections.

        Position Args: 
		label (str): The label to display to the user
		options (List[AbstraOption]): The options to display to the user, eg. ['Option 1', 'Option 2'] or [{'label': 'Option 1', 'value': '1'}, {'label': 'Option 2', 'value': '2'}]
		
        Keyword Args: 
		initial_value (str): The initial value to display to the user. Defaults to None.
		min (int): The minimum number of items that must be selected. Defaults to 0.
		max (int): The maximum number of items that can be selected. Defaults to the number of options.
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The selected values
    """

    
  
    
    key = kwargs.pop("key", label)
  
    self.widgets.append(ChecklistInput(key, label, options, **kwargs))
    return self

    
  def read_cnpj(self, label: str, **kwargs):
    
    """CNPJ number input with validation and masking.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (str): The initial value to display to the user. Defaults to None.
		invalid_message (str): The message to display when the user enters an invalid value. Defaults to None.
		placeholder (str): The placeholder text to display to the user. Defaults to "00.000.000/0001-00".
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The value entered by the user
    """

    
  
    
    key = kwargs.pop("key", label)
  
    self.widgets.append(CnpjInput(key, label, **kwargs))
    return self

    
  def read_code(self, label: str, **kwargs):
    
    """Accepts code input with syntax highlighting

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (str): The initial value to display to the user. Defaults to "".
		language (str): The programming language. Defaults to None.
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The value entered by the user
    """

    
  
    
    key = kwargs.pop("key", label)
  
    self.widgets.append(CodeInput(key, label, **kwargs))
    return self

    
  def read_cpf(self, label: str, **kwargs):
    
    """CPF number input with validation and masking.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (str): The initial value to display to the user. Defaults to "".
		invalid_message (str): The message to display when the user enters an invalid value. Defaults to None.
		placeholder (str): The placeholder text to display to the user. Defaults to "000.000.000-00".
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The value entered by the user
    """

    
  
    
    key = kwargs.pop("key", label)
  
    self.widgets.append(CpfInput(key, label, **kwargs))
    return self

    
  def read_currency(self, label: str, **kwargs):
    
    """Enables monetary value input with currency formatting.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (str): The initial value to display to the user. Defaults to 0.
		placeholder (str): The placeholder text to display to the user. Defaults to "".
		min (float): The minimum value allowed, eg. "0". Defaults to None.
		max (float): The maximum value allowed, eg. "100". Defaults to None.
		currency (str): The currency to display to the user, eg. "USD", "BRL, "EUR", "GBP". Defaults to "USD".
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The value entered by the user
    """

    
  
    
    key = kwargs.pop("key", label)
  
    self.widgets.append(CurrencyInput(key, label, **kwargs))
    return self

    
  def read_custom(self, html_body: str, **kwargs):
    
    """Allows for a fully customizable widget with custom HTML, CSS, and JS.

        Position Args: 
		html_body (str): The HTML body content
		
        Keyword Args: 
		initial_value (Any): The initial value to be stored in custom widget state.
		label (str): The label to display to the user
		html_head (str): The HTML head content
		css (str): The widget's CSS
		js (str): The widget's JavaScript
		height (int): The widget's height
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The custom response
    """

    
  
    
    default_key = kwargs.get("label") or self.__get_next_result_key()
    key = kwargs.pop("key", default_key)
    
    self.widgets.append(CustomInput(key, html_body, **kwargs))
    return self

    
  def read_date(self, label: str, **kwargs):
    
    """Provides a date picker with calendar interface.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (Union[datetime.date, time.struct_time, str]): The initial value to display to the user. Defaults to None.
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The value entered by the user
    """

    
  
    
    key = kwargs.pop("key", label)
  
    self.widgets.append(DateInput(key, label, **kwargs))
    return self

    
  def read_dropdown(self, label: str, options: List[AbstraOption], **kwargs):
    
    """Provides a dropdown menu for single or multiple selections.

        Position Args: 
		label (str): The label to display to the user
		options (List[AbstraOption]): The options to display to the user, eg. ['Option 1', 'Option 2'] or [{'label': 'Option 1', 'value': '1'}, {'label': 'Option 2', 'value': '2'}]
		
        Keyword Args: 
		multiple (bool): Whether the user can select multiple options. Defaults to False.
		initial_value (str or list): The initial value to display to the user. Defaults to [].
		placeholder (str): The placeholder text to display to the user. Defaults to "Choose an option".
		min (number): The minimal amount of options that should be selected. Defaults to None.
		max (number): The maximum amount of options that should be selected. Defaults to None.
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The value selected by the user
    """

    
  
    
    key = kwargs.pop("key", label)
  
    self.widgets.append(DropdownInput(key, label, options, **kwargs))
    return self

    
  def read_email(self, label: str, **kwargs):
    
    """ Collects and validates an email address.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (str): The initial value to display to the user. Defaults to "".
		placeholder (str): The placeholder text to display to the user. Defaults to "".
		invalid_email_message (str): Invalid e-mail message
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The value entered by the user
    """

    
  
    
    key = kwargs.pop("key", label)
  
    self.widgets.append(EmailInput(key, label, **kwargs))
    return self

    
  def read_file(self, label: str, **kwargs):
    
    """Enables file upload via file explorer.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (str): The initial value to display to the user. Defaults to "".
		multiple (bool): Whether the user will be allowed to upload multiple files. Defaults to False.
		max_file_size (float): Maximum size allowed to be transfered in total in MB.
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          A dictionary contains the file uploaded by the user, represented as FileResponse(file: TemporaryFile, content: bytes). If the multiple flag is set as True, it might contain a list of FileResponses.
    """

    
  
    
    key = kwargs.pop("key", label)
  
    self.widgets.append(FileInput(key, label, **kwargs))
    return self

    
  def read_image(self, label: str, **kwargs):
    
    """Allows for image file uploads with preview and validation.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (str): The initial value to display to the user. Defaults to "".
		multiple (bool): Whether the user will be allowed to upload multiple files. Defaults to False.
		max_file_size (float): Maximum size allowed to be transfered in total in MB.
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          A dict containing the image file uploaded by the user ({"file": file, "url": str, "content": bytes}) or a list of images in case of multiple flag set as True
    """

    
  
    
    key = kwargs.pop("key", label)
  
    self.widgets.append(ImageInput(key, label, **kwargs))
    return self

    
  def read_list(self, item_schema: Any, **kwargs):
    
    """Collects a dynamic list of values based on a specified schema.

        Position Args: 
		item_schema (Any): The schema for the items of the list
		
        Keyword Args: 
		initial_value (array): ''
		min (float): Min value accepted by the input. Defaults to None.
		max (float): Max value accepted by the input. Defaults to None.
		add_button_text (str): ''
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The values entered by the user
    """

    
  
    
    default_key = self.__get_next_result_key()
    key = kwargs.pop("key", default_key)
    
    self.widgets.append(ListInput(key, item_schema, **kwargs))
    return self

    
  def read_multiple_choice(self, label: str, options: List[AbstraOption], **kwargs):
    
    """Offers multiple choice selections with single or multi-select options.

        Position Args: 
		label (str): The label to display to the user
		options (List[AbstraOption]): The options to display to the user, eg. ['Option 1', 'Option 2'] or [{'label': 'Option 1', 'value': '1'}, {'label': 'Option 2', 'value': '2'}]
		
        Keyword Args: 
		multiple (bool): Whether the user can select multiple options. Defaults to False.
		min (number): The minimal amount of options that should be selected. Defaults to None.
		max (number): The maximum amount of options that should be selected. Defaults to None.
		initial_value ([]): The initial value of the selection. Defaults to [].
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The values/value selected by the user
    """

    
  
    
    key = kwargs.pop("key", label)
  
    self.widgets.append(MultipleChoiceInput(key, label, options, **kwargs))
    return self

    
  def read_nps(self, label: str, **kwargs):
    
    """Captures Net Promoter Score feedback with a 0-10 rating scale.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		min (int): Min value accepted by the input. Defaults to 0.
		max (int): Max value accepted by the input. Defaults to 10.
		min_hint (str): Text to display next to the min value. Defaults to "Not at all likely".
		max_hint (str): Text to display next to the max value. Defaults to "Extremely likely".
		initial_value (str): The initial value to display to the user. Defaults to None.
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The value entered by the user
    """

    
  
    
    key = kwargs.pop("key", label)
  
    self.widgets.append(NpsInput(key, label, **kwargs))
    return self

    
  def read_number(self, label: str, **kwargs):
    
    """Collects numeric input with optional min/max limits.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (str): The initial value to display to the user. Defaults to 0.
		placeholder (str): The placeholder text to display to the user. Defaults to "".
		min (float): Min value accepted by the input. Defaults to None.
		max (float): Max value accepted by the input. Defaults to None.
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The value entered by the user
    """

    
  
    
    key = kwargs.pop("key", label)
  
    self.widgets.append(NumberInput(key, label, **kwargs))
    return self

    
  def read_number_slider(self, label: str, **kwargs):
    
    """Provides a slider for selecting numerical values within a range.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (float): The initial value to display to the user. Defaults to 0.
		min (float): Min value accepted by the input. Defaults to None.
		max (float): Max value accepted by the input. Defaults to None.
		step (float): The value to be incremented or decremented while using the input button. Defaults to None.
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The value entered by the user
    """

    
  
    
    key = kwargs.pop("key", label)
  
    self.widgets.append(NumberSliderInput(key, label, **kwargs))
    return self

    
  def read_pandas_row_selection(self, df: PandasDataFrame, **kwargs):
    
    """Enables selection of rows from a displayed pandas DataFrame table.

        Position Args: 
		df (PandasDataFrame): The pandas dataframe to be displayed
		
        Keyword Args: 
		display_index (bool): Whether to show a index column. Defaults to False.
		label (str): The label to display to the user
		initial_value (list): The initial value of the selection. Defaults to []
		multiple (bool): Whether the user will be allowed to select multiple rows. Defaults to False.
		min (number): The minimal amount of options that should be selected. Defaults to None.
		max (number): The maximum amount of options that should be selected. Defaults to None.
		page_size (number): The number of rows to display per page. Defaults to 10.
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The list of selected rows
    """

    
  
    
    default_key = kwargs.get("label") or self.__get_next_result_key()
    key = kwargs.pop("key", default_key)
    
    self.widgets.append(PandasRowSelectionInput(key, df, **kwargs))
    return self

    
  def read_password(self, label: str, **kwargs):
    
    """A password field, including strength validation options.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		placeholder (str): The placeholder text to display to the user. Defaults to "".
		lowercase_required (bool): Whether the input must have at least one lowercase character. Defaults to True.
		uppercase_required (bool): Whether the input must have at least one uppercase character. Defaults to True.
		special_required (bool): Whether the input must have at least one special character. Defaults to True.
		digit_required (bool): Whether the input must have at least one digit. Defaults to True.
		min_length (int): Minimum length of the password. Defaults to 8.
		max_length (int): Maximum length of the password. Defaults to None.
		size (int): Size of the password. Defaults to None.
		pattern (str): A regex pattern for the accepted password. Defaults to None.
		autocomplete (str): The autocomplete HTML attribute. Defaults to "current-password".
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The value entered by the user
    """

    
  
    
    key = kwargs.pop("key", label)
  
    self.widgets.append(PasswordInput(key, label, **kwargs))
    return self

    
  def read_phone(self, label: str, **kwargs):
    
    """A phone inputs with country code and national number.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (dict): The initial value to display to the user. It contains two keys: 'country_code' (string with optional + sign or number) and 'national_number' (str). Ex: {'country_code': '55', 'national_number': '21999990000'}.
		placeholder (str): The placeholder text to display in the national number input. Defaults to "".
		invalid_message (str): The message to display when the input is invalid
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          A dict containing the value entered by the user ({"country_code": str, "national_number": str})
    """

    
  
    
    key = kwargs.pop("key", label)
  
    self.widgets.append(PhoneInput(key, label, **kwargs))
    return self

    
  def read_rating(self, label: str, **kwargs):
    
    """Collects user feedback with emoji icons.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (int): The initial value to display to the user. Defaults to 0.
		max (float): Max value accepted by the input. Defaults to None.
		char (str): Which char should be displayed as icon?
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The value entered by the user
    """

    
  
    
    key = kwargs.pop("key", label)
  
    self.widgets.append(RatingInput(key, label, **kwargs))
    return self

    
  def read_richtext(self, label: str, **kwargs):
    
    """Offers a rich text editor for formatted textual input.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (str): The initial value to display to the user. Defaults to "".
		placeholder (str): The placeholder text to display to the user. Defaults to "Your rich text here".
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The rich text value entered by the user
    """

    
  
    
    key = kwargs.pop("key", label)
  
    self.widgets.append(RichTextInput(key, label, **kwargs))
    return self

    
  def read_tag(self, label: str, **kwargs):
    
    """Enables input of tags or keywords with autocomplete suggestions.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (list): The initial value to display to the user. Defaults to [].
		placeholder (str): The placeholder text to display to the user. Defaults to "".
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The values entered by the user
    """

    
  
    
    key = kwargs.pop("key", label)
  
    self.widgets.append(TagInput(key, label, **kwargs))
    return self

    
  def read(self, label: str, **kwargs):
    
    """Collects plain text input with customizable placeholders and validation.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (str): The initial value to display to the user. Defaults to "".
		placeholder (str): The placeholder text to display to the user. Defaults to "Placeholder".
		mask (str): A mask to apply to the input. Defaults to None.
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The value entered by the user
    """

    
  
    
    key = kwargs.pop("key", label)
  
    self.widgets.append(TextInput(key, label, **kwargs))
    return self

    
  def read_textarea(self, label: str, **kwargs):
    
    """Provides a text area for multi-line text input.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (str): The initial value to display to the user. Defaults to "".
		placeholder (str): The placeholder text to display to the user. Defaults to "".
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The value entered by the user
    """

    
  
    
    key = kwargs.pop("key", label)
  
    self.widgets.append(TextareaInput(key, label, **kwargs))
    return self

    
  def read_time(self, label: str, **kwargs):
    
    """Allows selection of a specific time.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (str or datetime.time): The initial value to display to the user. Defaults to "00:00".
		format (str): Whether the input is in the format 24hs or AM/PM. Defaults to "24hs".
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          A datetime.time object representing the value entered by the user
    """

    
  
    
    key = kwargs.pop("key", label)
  
    self.widgets.append(TimeInput(key, label, **kwargs))
    return self

    
  def read_toggle(self, label: str, **kwargs):
    
    """Offers a toggle switch between two defined options.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		on_text (str): Text of On Toggle option
		off_text (str): Text of Off Toggle option
		initial_value (bool): Initial value of the toggle
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          The toggle value
    """

    
  
    
    key = kwargs.pop("key", label)
  
    self.widgets.append(ToggleInput(key, label, **kwargs))
    return self

    
  def read_video(self, label: str, **kwargs):
    
    """Facilitates video file uploads with preview and size validation.

        Position Args: 
		label (str): The label to display to the user
		
        Keyword Args: 
		initial_value (Union[str, io.IOBase]): The initial value to display to the user. Defaults to "".
		multiple (bool): Whether the user will be allowed to upload multiple files. Defaults to False.
		max_file_size (float): Maximum size allowed to be transfered in total in MB.
		disabled (bool): whether the input is disabled. Defaults to False.
		required (Union[bool, str]): Whether the input is required or not eg. "this field is required". Defaults to True.
		hint (str): A tooltip displayed to the user. Defaults to None.
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        Returns:
          A dict containing the video uploaded by the user ({"file": file, "url": str, "content": bytes}) or a list of videos in case of multiple flag set as True
    """

    
  
    
    key = kwargs.pop("key", label)
  
    self.widgets.append(VideoInput(key, label, **kwargs))
    return self

    
  def display_file(self, file: Union[str, io.IOBase], **kwargs):
    
    """Show a button for the user to download a file.

        Position Args: 
		file (Union[str, io.IOBase]): The file to download
		
        Keyword Args: 
		download_text (str): The text to display on the button that will download the file. Defaults to "Download here".
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        
    """

    
  
    
    self.widgets.append(FileOutput(file, **kwargs))
    return self

    
  def display_html(self, html: str, **kwargs):
    
    """Simple HTML embed widget.

        Position Args: 
		html (str): The html snippet to display to the user
		
        Keyword Args: 
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        
    """

    
  
    
    self.widgets.append(HtmlOutput(html, **kwargs))
    return self

    
  def display_iframe(self, url_or_html: str, **kwargs):
    
    """Embed content using an iframe.

        Position Args: 
		url_or_html (str): The link to the document or the own document to display to the user
		
        Keyword Args: 
		width (int): The width of the iframe. Defaults to "800".
		height (int): The height of the iframe. Defaults to "600".
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        
    """

    
  
    
    self.widgets.append(IframeOutput(url_or_html, **kwargs))
    return self

    
  def display_image(self, image: Union[str, io.IOBase], **kwargs):
    
    """Show an image to the user

        Position Args: 
		image (Union[str, io.IOBase]): The image to display to the user
		
        Keyword Args: 
		subtitle (str): The subtitle of the image. Defaults to "".
		label (str): The label to display to the user
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        
    """

    
  
    
    self.widgets.append(ImageOutput(image, **kwargs))
    return self

    
  def display_latex(self, text: str, **kwargs):
    
    """Display a latex formula to the user.

        Position Args: 
		text (str): The latex formula to display to the user
		
        Keyword Args: 
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        
    """

    
  
    
    self.widgets.append(LatexOutput(text, **kwargs))
    return self

    
  def display_link(self, link_url: str, **kwargs):
    
    """Display a link to the user.

        Position Args: 
		link_url (str): The url of the link to display to the user
		
        Keyword Args: 
		link_text (str): The text to display on the link. Defaults to "Click here".
		same_tab (bool): Whether to open the link in the same tab or not. Defaults to False.
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        
    """

    
  
    
    self.widgets.append(LinkOutput(link_url, **kwargs))
    return self

    
  def display_markdown(self, text: str, **kwargs):
    
    """Show markdown formatted text to the user

        Position Args: 
		text (str): The formatted text to display to the user
		
        Keyword Args: 
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        
    """

    
  
    
    self.widgets.append(MarkdownOutput(text, **kwargs))
    return self

    
  def display_pandas(self, df: PandasDataFrame, **kwargs):
    
    """Display a pandas dataframe to the user

        Position Args: 
		df (PandasDataFrame): The dataframe to display to the user
		
        Keyword Args: 
		display_index (bool): Whether to show a index column. Defaults to False.
		label (str): The label to display to the user
		editable (bool): Whether the user will be allowed to edit the cells. Defaults to False.
		page_size (number): The number of rows to display per page. Defaults to 10.
		actions (list): Actions that can be triggered by table rows
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        
    """

    
  
    
    self.widgets.append(PandasOutput(df, **kwargs))
    return self

    
  def display_plotly(self, fig: PlotlyFigure, **kwargs):
    
    """Displays a Plotly figure.

        Position Args: 
		fig (PlotlyFigure): The figure to display to the user
		
        Keyword Args: 
		label (str): The label to display to the user
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        
    """

    
  
    
    self.widgets.append(PlotlyOutput(fig, **kwargs))
    return self

    
  def display_progress(self, current: float, total: float, **kwargs):
    
    """Displays a progress bar to the user.

        Position Args: 
		current (float): The progress being made. Defaults to 50.
		total (float): Total progress. Defaults to 100.
		
        Keyword Args: 
		text (str): The text displayed with this progress step. Defaults to "".
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        
    """

    
  
    
    self.widgets.append(ProgressOutput(current, total, **kwargs))
    return self

    
  def display(self, text: str, **kwargs):
    
    """"Text display" adds textual information into the app.

        Position Args: 
		text (str): The text to display to the user
		
        Keyword Args: 
		size (str): The size of the text: 'small', 'medium', 'large'
		end_program (bool): Whether the program should end after the widget is shown. Defaults to False.
		full_width (bool): Whether the input should use full screen width. Defaults to False.
		button_text (str): What text to display on the button when the widget is not part of a Page. Defaults to 'Next'.
		
        
    """

    
  
    
    self.widgets.append(TextOutput(text, **kwargs))
    return self

    
  def serialize_value(self):
    serialized = {}
    for widget in self.get_input_widgets():
        if isinstance(widget, Reactive):
            serialized.update(widget.serialize_value())
        else:
            serialized[widget.key] = widget.serialize_value()
    return serialized

  input = read



from unittest import TestCase
from unittest.mock import ANY

from ...widgets.library.DropdownInput import DropdownInput


class TestHotReloadLocalModules(TestCase):
    maxDiff = None

    def test_initial_render_dict_list(self):
        widget = DropdownInput(
            key="dropdown_key_1",
            label="Dropdown Label",
            options=[{"label": "Option 1", "value": "option_1"}],
        )
        result = widget.render({})

        # option values are generated automatically
        self.assertEqual(
            result,
            {
                "type": "dropdown-input",
                "key": "dropdown_key_1",
                "label": "Dropdown Label",
                "options": [{"label": "Option 1", "value": ANY}],
                "hint": None,
                "multiple": False,
                "placeholder": "",
                "value": [],
                "required": True,
                "fullWidth": False,
                "disabled": False,
                "min": None,
                "max": None,
                "errors": [],
            },
        )

    def test_initial_render_str_list(self):
        widget = DropdownInput(
            key="dropdown_key_1",
            label="Dropdown Label",
            options=["Option 1", "Option 2"],
        )

        result = widget.render({})

        self.assertEqual(
            result,
            {
                "type": "dropdown-input",
                "key": "dropdown_key_1",
                "label": "Dropdown Label",
                "options": [
                    {"label": "Option 1", "value": ANY},
                    {"label": "Option 2", "value": ANY},
                ],
                "hint": None,
                "multiple": False,
                "placeholder": "",
                "value": [],
                "required": True,
                "fullWidth": False,
                "disabled": False,
                "errors": [],
                "min": None,
                "max": None,
            },
        )

    def test_render_setting_value(self):
        widget = DropdownInput(
            key="dropdown_key_1",
            label="Dropdown Label",
            options=["Option 1", "Option 2"],
        )

        initial_render = widget.render({})

        user_sent_value = [initial_render["options"][1]["value"]]
        parsed_value = widget.parse_value(user_sent_value)
        widget.set_value(parsed_value)

        result = widget.render({})

        self.assertEqual(
            result,
            {
                "type": "dropdown-input",
                "key": "dropdown_key_1",
                "label": "Dropdown Label",
                "options": [
                    {
                        "label": "Option 1",
                        "value": initial_render["options"][0]["value"],
                    },
                    {
                        "label": "Option 2",
                        "value": initial_render["options"][1]["value"],
                    },
                ],
                "hint": None,
                "multiple": False,
                "placeholder": "",
                "value": [initial_render["options"][1]["value"]],
                "required": True,
                "fullWidth": False,
                "disabled": False,
                "errors": [],
                "min": None,
                "max": None,
            },
        )

    def test_min_error(self):
        widget = DropdownInput(
            key="dropdown_key_1",
            label="Dropdown Label",
            options=["Option 1", "Option 2", "Option 3"],
            multiple=True,
            min=2,
        )

        initial_render = widget.render({})
        user_sent_value = [initial_render["options"][1]["value"]]
        parsed_value = widget.parse_value(user_sent_value)
        widget.set_value(parsed_value, set_errors=True)

        result = widget.render({})

        self.assertEqual(
            result,
            {
                "type": "dropdown-input",
                "key": "dropdown_key_1",
                "label": "Dropdown Label",
                "options": [
                    {
                        "label": "Option 1",
                        "value": initial_render["options"][0]["value"],
                    },
                    {
                        "label": "Option 2",
                        "value": initial_render["options"][1]["value"],
                    },
                    {
                        "label": "Option 3",
                        "value": initial_render["options"][2]["value"],
                    },
                ],
                "hint": None,
                "multiple": True,
                "placeholder": "",
                "value": [initial_render["options"][1]["value"]],
                "required": True,
                "fullWidth": False,
                "disabled": False,
                "errors": ["i18n_error_min_list"],
                "min": 2,
                "max": None,
            },
        )

    def test_max_error(self):
        widget = DropdownInput(
            key="dropdown_key_1",
            label="Dropdown Label",
            options=["Option 1", "Option 2", "Option 3"],
            multiple=True,
            max=1,
        )

        initial_render = widget.render({})
        user_sent_value = [
            initial_render["options"][0]["value"],
            initial_render["options"][1]["value"],
        ]
        parsed_value = widget.parse_value(user_sent_value)
        widget.set_value(parsed_value, set_errors=True)

        result = widget.render({})

        self.assertEqual(
            result,
            {
                "type": "dropdown-input",
                "key": "dropdown_key_1",
                "label": "Dropdown Label",
                "options": [
                    {
                        "label": "Option 1",
                        "value": initial_render["options"][0]["value"],
                    },
                    {
                        "label": "Option 2",
                        "value": initial_render["options"][1]["value"],
                    },
                    {
                        "label": "Option 3",
                        "value": initial_render["options"][2]["value"],
                    },
                ],
                "hint": None,
                "multiple": True,
                "placeholder": "",
                "value": [
                    initial_render["options"][0]["value"],
                    initial_render["options"][1]["value"],
                ],
                "required": True,
                "fullWidth": False,
                "disabled": False,
                "errors": ["i18n_error_max_list"],
                "min": None,
                "max": 1,
            },
        )



from ..widget_base import Output
from typing import Union
from ..file_utils import convert_file
import io


class FileOutput(Output):
    type = 'file-output'

    def __init__(self, file: Union[str, io.IOBase], **kwargs):
        self.set_props(dict(file=file, **kwargs))

    def set_props(self, props):
        self.file = convert_file(props.get('file', ''))
        self.download_text = props.get('download_text', 'Download')
        self.full_width = props.get('full_width', False)

    def render(self, ctx: dict):
        return {'type': self.type, 'fileUrl': self.file, 'downloadText':
            self.download_text, 'fullWidth': self.full_width}



from typing import List

import flask

from ...contracts_generated import CommonUserRoles
from ...jwt_auth import USER_AUTH_HEADER_KEY, UserClaims
from ...repositories import users_repository
from ...repositories.project.project import ProjectRepository
from ...repositories.users import UsersRepository
from ...usage import usage


class UserController:
    def __init__(
        self,
        users_repository: UsersRepository,
    ) -> None:
        self.users_repository = users_repository

    def roles(self, claims: UserClaims):
        return self.users_repository.get_user_roles(claims.email)

    def sign_up(self, claims: UserClaims) -> bool:
        policy = self.get_sign_up_policy()
        if not policy.allow(claims.email):
            return False

        if self.users_repository.get_user(claims.email) is not None:
            return True

        return self.users_repository.insert_user(claims.email, [])

    def list_users(self):
        return self.users_repository.list_users()

    def insert_user(self, email: str, roles: CommonUserRoles):
        self.users_repository.insert_user(email, roles)

    def delete_user(self, email: str):
        self.users_repository.delete_user(email)

    def update_user(self, email: str, roles: CommonUserRoles):
        self.users_repository.update_user(email, roles)

    def update_sign_up_policy(self, email_patterns: List[str]):
        project = ProjectRepository.load()
        project.signup_policy.update({"email_patterns": email_patterns})
        ProjectRepository.save(project)

    def get_sign_up_policy(self):
        project = ProjectRepository.load()
        return project.signup_policy


def get_editor_bp():
    controller = UserController(users_repository)
    bp = flask.Blueprint("users", __name__)

    @bp.get("/")
    @usage
    def _list_users():
        return controller.list_users()

    @bp.post("/")
    @usage
    def _insert_user():
        data = flask.request.json
        if data is None or "email" not in data:
            flask.abort(400)

        email = data["email"]
        roles = data.get("roles", [])
        controller.insert_user(email, roles)
        return {"success": True}

    @bp.patch("/<email>")
    @usage
    def _update_user(email):
        data = flask.request.json
        if data is None or "roles" not in data:
            flask.abort(400)
        roles = data.get("roles", [])
        controller.update_user(email, roles)
        return {"success": True}

    @bp.delete("/<email>")
    @usage
    def _delete_user(email):
        controller.delete_user(email)
        return {"success": True}

    @bp.put("/signup-policy")
    @usage
    def _update_sign_up_policy():
        data = flask.request.json
        if data is None or "email_patterns" not in data:
            flask.abort(400)

        email_patterns = data.get("email_patterns", [])
        controller.update_sign_up_policy(email_patterns)
        return {"success": True}

    @bp.get("/signup-policy")
    @usage
    def _get_sign_up_policy():
        return controller.get_sign_up_policy().as_dict

    return bp


def get_player_bp():
    controller = UserController(users_repository)
    bp = flask.Blueprint("playerUsers", __name__)

    @bp.get("/my-roles")
    def _get_user_roles():
        auth = flask.request.headers.get(USER_AUTH_HEADER_KEY)
        if auth is None:
            return []

        claims = UserClaims.from_jwt(auth.split(" ")[1])
        if claims is None or claims.email is None:
            return []

        return controller.roles(claims)

    @bp.post("/sign-up")
    def _signup_user():
        auth = flask.request.headers.get(USER_AUTH_HEADER_KEY)
        if auth is None:
            flask.abort(401)

        claims = UserClaims.from_jwt(auth.split(" ")[1])
        if claims is None or claims.email is None:
            flask.abort(401)

        return {"success": controller.sign_up(claims)}

    return bp



from gunicorn.arbiter import Arbiter
from gunicorn.workers.base import Worker

from ..utils.environment import set_SERVER_UUID, set_WORKER_UUID


# Called just before the master process is initialized.
def on_starting(server: Arbiter):
    set_SERVER_UUID()


# Called just before a worker is forked.
def pre_fork(server: Arbiter, worker: Worker):
    set_WORKER_UUID()



import io
import pathlib
import tempfile
import typing as t

from .apis import get_random_filepath, internal_path, upload_file


def convert_file(file: t.Union[str, io.IOBase, pathlib.Path]) -> str:
    if not file:
        return ""

    if isinstance(file, (io.IOBase, pathlib.Path)):
        return upload_file(file)

    if isinstance(file, str):
        # URL or base64 encoded string
        if file.startswith("http") or file.startswith("data:"):
            return file

        # path to file
        return upload_file(open(file, "rb"))

    # PILImage. TODO: check with isinstance without external dependency
    if hasattr(file, "save"):
        _, file_path = get_random_filepath()
        file.save(str(file_path))
        return upload_file(open(file_path, "rb"))

    # FileResponse. TODO: check with isinstance without circular import
    if hasattr(file, "file"):
        return upload_file(file.file)

    raise ValueError(f"Cannot convert {type(file)}")


def download_file(
    url: str,
) -> t.Union[io.BufferedReader, tempfile._TemporaryFileWrapper]:
    import requests

    if url.startswith("http://") or url.startswith("https://"):
        f = tempfile.NamedTemporaryFile()
        with requests.get(url, stream=True) as r:
            r.raise_for_status()
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)
        f.seek(0)

        return f

    elif url.startswith("/_files/"):
        name = url[len("/_files/") :]
        path = internal_path(name)
        return open(path, "rb")

    raise ValueError(f"Cannot download {url}")



import os
import pathlib
import sys
import typing

from .utils import get_free_port
from .utils.environment import DEFAULT_PORT


class SettingsController(object):
    _instance: typing.Optional["SettingsController"] = None
    _root_path: typing.Optional[pathlib.Path] = None
    _server_port: typing.Optional[int] = None

    def __new__(cls):
        if not cls._instance:
            cls._instance = super(SettingsController, cls).__new__(cls)
        return cls._instance

    @staticmethod
    def set_root_path(root_path_str: str):
        root_path = pathlib.Path(root_path_str).absolute()
        root_path.mkdir(exist_ok=True, parents=True)
        os.chdir(root_path)

        if root_path not in sys.path:
            sys.path.append(str(root_path))

        SettingsController._root_path = root_path

    @staticmethod
    def set_server_port(server_port: typing.Union[int, str, None]):
        port = 3000
        if isinstance(server_port, int):
            port = server_port
        elif isinstance(server_port, str):
            port = int(server_port)
        elif DEFAULT_PORT:
            port = int(DEFAULT_PORT)

        free_port = get_free_port(port)
        SettingsController._server_port = free_port

    @property
    def root_path(self) -> pathlib.Path:
        if SettingsController._root_path is None:
            raise Exception("You must set the root path before using it")
        return SettingsController._root_path

    @property
    def server_port(self) -> int:
        if SettingsController._server_port is None:
            raise Exception("You must set the server before using it")
        return SettingsController._server_port


Settings = SettingsController()



from ..logger import AbstraLogger
from ..server.apps import get_cloud_app
from ..server.controller.main import MainController
from ..settings import SettingsController
from ..stdio_monkey_patch import override_stdio
from ..utils.environment import (
    DEFAULT_PORT,
    THREADS,
    WORKER_CLASS,
    WORKER_TEMP_DIR,
    WORKERS,
)
from .application import CustomApplication
from .hooks import on_starting, pre_fork


def run():
    AbstraLogger.init("cloud")
    override_stdio(print_exceptions=True)
    SettingsController.set_root_path(".")  # TODO: use CWD
    SettingsController.set_server_port(DEFAULT_PORT)

    options = {
        "workers": WORKERS,
        "threads": THREADS,
        "worker_class": WORKER_CLASS,
        "worker_tmp_dir": WORKER_TEMP_DIR,
        "bind": f":{DEFAULT_PORT or 8002}",
        "on_starting": on_starting,
        "pre_fork": pre_fork,
    }

    controller = MainController()
    app = get_cloud_app(controller)
    CustomApplication(app, options).run()


if __name__ == "__main__":
    run()



from ..widget_base import Output
from typing import Any
import json
DEFAULT_PAGE_SIZE = 10


class PandasOutput(Output):
    type = 'pandas-output'

    def __init__(self, df: Any=None, **kwargs):
        self.set_props(dict(df=df, **kwargs))

    def set_props(self, props):
        self.df = props.get('df', '')
        self.full_width = props.get('full_width', False)
        self.display_index = props.get('display_index', False)
        self.label = props.get('label', '')
        self.actions = props.get('actions', [])
        self.filterable = props.get('filterable', False)
        self.editable = props.get('editable', False)
        self.page_size = props.get('page_size', DEFAULT_PAGE_SIZE)

    def serialize_table(self):
        if self.df is None:
            import pandas as pd
            df = pd.DataFrame({'change the': [1, 2, 3], 'df property': [4, 
                5, 6]})
            df_json = df.to_json(orient='table')
            if not isinstance(df_json, str):
                raise Exception('df.to_json() did not return a string')
            serialized = json.loads(df_json)
            del serialized['schema']['pandas_version']
            return serialized
        serialized = json.loads(self.df.to_json(orient='table'))
        del serialized['schema']['pandas_version']
        return serialized

    def render(self, ctx: dict):
        return {'type': self.type, 'table': self.serialize_table(),
            'fullWidth': self.full_width, 'displayIndex': self.
            display_index, 'label': self.label, 'actions': self.actions,
            'filterable': self.filterable, 'editable': self.editable,
            'pageSize': self.page_size}



from typing import List
from ..widget_base import Input, NumberValueHandler


class NpsInput(Input):
    type = 'nps-input'

    def __init__(self, key: str, label: str, **kwargs):
        super().__init__(key)
        self.set_props(dict(label=label, **kwargs))

    def set_props(self, props):
        self.label = props.get('label', 'Label')
        self.value = props.get('initial_value', None)
        self.required = props.get('required', True)
        self.min = props.get('min', 0)
        self.max = props.get('max', 10)
        self.min_hint = props.get('min_hint', 'Not at all likely')
        self.max_hint = props.get('max_hint', 'Extremely likely')
        self.hint = props.get('hint', None)
        self.full_width = props.get('full_width', False)
        self.disabled = props.get('disabled', False)
        self.number_value_handler = NumberValueHandler(min=self.min, max=
            self.max)

    def render(self, ctx: dict):
        return {'type': self.type, 'key': self.key, 'label': self.label,
            'min': self.min, 'max': self.max, 'minHint': self.min_hint,
            'maxHint': self.max_hint, 'value': self.serialize_value(),
            'required': self.required, 'hint': self.hint, 'fullWidth': self
            .full_width, 'disabled': self.disabled, 'errors': self.errors}

    def validate(self) ->List[str]:
        return self.number_value_handler.validate(self.value)

    def serialize_value(self) ->int:
        return self.value



import unittest

from abstra.forms import ListItemSchema, Page


class TestPage(unittest.TestCase):
    def test_reactive_list_dropdown(self):
        schema = ListItemSchema().read_dropdown("dropdown", ["foo", "bar"])

        def render(ctx):
            return Page().read_list(schema, key="list", min=1)

        page = Page().reactive(render)

        rendered = page.render({})

        bar_value = rendered[0]["schemas"][0][0]["options"][1]["value"]

        browser_input = {"list": [{"dropdown": [bar_value]}]}
        parsed_value = page.parse_value(browser_input)

        self.assertEqual(parsed_value, {"list": [{"dropdown": "bar"}]})



import os

from .settings import Settings
from .utils.dot_abstra import CREDENTIALS_FILE


def get_credentials():
    if os.getenv("ABSTRA_API_TOKEN"):
        return os.getenv("ABSTRA_API_TOKEN")

    credentials_path = Settings.root_path.joinpath(CREDENTIALS_FILE)
    if not credentials_path.exists():
        return None

    return credentials_path.read_text(encoding="utf-8").strip()


def delete_credentials():
    credentials_path = Settings.root_path.joinpath(CREDENTIALS_FILE)

    if credentials_path.exists():
        credentials_path.unlink()


def set_credentials(token: str):
    credentials_path = Settings.root_path.joinpath(CREDENTIALS_FILE)
    credentials_path.parent.mkdir(exist_ok=True)

    credentials_path.write_text(token, encoding="utf-8")


def resolve_headers():
    credentials = get_credentials()

    if not credentials:
        return None

    return {"Api-Authorization": f"Bearer {credentials}"}



from datetime import date
from unittest import TestCase

from . import check_is_url, deserialize, is_serializable, nested_get, serialize


class SerializeTest(TestCase):
    def test_serialize(self):
        obj = {"a": 1}
        result = serialize(obj)
        self.assertEqual(result, '{"a": 1}')

    def test_serialize_nan(self):
        obj = {"a": float("nan")}
        result = serialize(obj)
        self.assertEqual(result, '{"a": null}')

    def test_deserialize(self):
        obj = '{"a": 1}'
        result = deserialize(obj)
        self.assertEqual(result, {"a": 1})

    def test_is_serializable(self):
        obj = {"a": 1}
        result = is_serializable(obj)
        self.assertTrue(result)

    def test_is_not_serializable(self):
        obj = {"a": date.today()}
        result = is_serializable(obj)
        self.assertFalse(result)


class CheckIsUrlTest(TestCase):
    def test_is_url_https(self):
        url = "https://www.google.com"
        result = check_is_url(url)
        self.assertTrue(result)

    def test_is_url_http(self):
        url = "http://www.google.com"
        result = check_is_url(url)
        self.assertTrue(result)

    def test_is_not_url(self):
        url = "not a url"
        result = check_is_url(url)
        self.assertFalse(result)


class NestedGetTest(TestCase):
    def test_nested_get(self):
        obj = {"a": {"b": 1}}
        result = nested_get(obj, "a.b")
        self.assertEqual(result, 1)

    def test_nested_get_default_none(self):
        obj = {"a": {"b": 1}}
        result = nested_get(obj, "a.c")
        self.assertIsNone(result)



import json
from dataclasses import is_dataclass
from datetime import date, datetime
from uuid import UUID

from ..interface.sdk.forms.page_response import PageResponse
from ..interface.sdk.forms.step import StepsResponse
from ..utils import serialize
from ..widgets.response_types import FileResponse, PhoneResponse


def to_json_serializable(data):
    if isinstance(data, FileResponse):
        return data.url
    elif isinstance(data, PhoneResponse):
        return data.masked
    elif isinstance(data, PageResponse):
        return to_json_serializable(data.data)
    elif isinstance(data, StepsResponse):
        return to_json_serializable(data.acc)
    elif is_dataclass(data):
        return {k: to_json_serializable(v) for k, v in data.__dict__.items()}
    elif isinstance(data, (datetime, date)):
        return data.isoformat()
    elif isinstance(data, UUID):
        return str(data)
    elif isinstance(data, (set, list, tuple)):
        return [to_json_serializable(d) for d in data]
    elif isinstance(data, dict):
        return {k: to_json_serializable(v) for k, v in data.items()}
    return json.loads(serialize(data))



from .base_migration import Migration


class Migration002(Migration):
    @staticmethod
    def target_version() -> str:
        return "2.0"

    @staticmethod
    def source_version() -> str:
        return "1.0"

    def _add_visualization_settings(self) -> None:
        self.data["visualization"] = []

    def _transition_label_to_type(self) -> None:
        for stage_type in ["scripts", "forms", "hooks", "jobs"]:
            for stage in self.data[stage_type]:
                stage["transitions"] = [
                    {
                        "target_id": transition["target_id"],
                        "target_type": transition["target_type"],
                        "type": transition.get("type", f"{stage_type}:finished"),
                        "id": transition["id"],
                    }
                    for transition in stage["transitions"]
                ]

    def _add_control_stages(self) -> None:
        self.data["iterators"] = []
        self.data["conditions"] = []

    def _migrate(self) -> None:
        self._transition_label_to_type()
        self._add_visualization_settings()
        self._add_control_stages()



from typing import Sequence

from ...repositories.requirements import (
    RequirementRecommendation,
    RequirementsRepository,
)
from ..linter import LinterFix, LinterIssue, LinterRule


class AddMissingPackagesToRequirements(LinterFix):
    def __init__(self, requirement_recommendation: RequirementRecommendation):
        self.requirement_recommendation = requirement_recommendation

    def make_label(self):
        return f"Add {self.requirement_recommendation.requirement.name} to requirements.txt"

    def fix(self):
        requirements = RequirementsRepository.load()
        requirement_name = self.requirement_recommendation.requirement.name
        requirement_version = self.requirement_recommendation.requirement.version
        requirements.add(requirement_name, requirement_version)
        RequirementsRepository.save(requirements)


class PackageNotInRequirementsFound(LinterIssue):
    def __init__(self, recommendation: RequirementRecommendation) -> None:
        self.label = "\n".join(
            [
                f"You're importing the {recommendation.requirement.name}.",
                f"It is being used in {recommendation.reason_file}:{recommendation.reason_line}:",
                f"{recommendation.reason_code}",
                "But it's not in your requirements.txt file.",
            ]
        )
        self.fixes = [AddMissingPackagesToRequirements(recommendation)]


class MissingPackagesInRequirements(LinterRule):
    label = "All imported packages should be in requirements.txt"
    type = "bug"

    def find_issues(self) -> Sequence[LinterIssue]:
        recommendations = RequirementsRepository.get_recommendation()
        issues = [
            PackageNotInRequirementsFound(recommendation)
            for recommendation in recommendations
        ]
        return issues
